/* get_re
 * 
 * Reads an XRSL file and prints runtime environment names to stdout.
 *
 * Compile with:
 * gcc -Wall -I$GLOBUS_LOCATION/include/gcc32dbg -L$GLOBUS_LOCATION/lib \
 * -lglobus_rsl_gcc32dbg -o get_re get_re.c
 *
 * Leif Nixon <nixon@nsc.liu.se>
 *
 */

#include <stdio.h>

#include "globus_common.h"
#include "globus_rsl.h"

void die(char *msg) {
  fprintf(stderr, "get_re: %s\n", msg);
  exit(1);
}

void get_runtimeenvironments(char *rsl) {
  globus_rsl_t  *rsl_tree;
  char **params;
  int error, i;

  rsl_tree = globus_rsl_parse(rsl);
  if (!rsl_tree) die("couldn't parse XRSL");

  error = globus_rsl_param_get(rsl_tree,
			       GLOBUS_RSL_PARAM_MULTI_LITERAL,
			       "runtimeenvironment",
			       &params); 

  if (error) die("error in finding runtimeenvironment parameter");

  for (i=0; params[i]; i++) {
    printf("%s\n", params[i]);
  }
}

char *read_file(FILE *fp) {
  char *str, *s;
  int len = 0;

  str = malloc(4096);
  if (!str) die("couldn't allocate memory");
  s = str;

  while(fgets(s, 4096, fp)) {
    len += strlen(s);
    str = realloc(str, len+4096);
    if (!str) die("couldn't allocate memory");
    s = str+len;
  }

  return str;
}

int main(int argc, char **argv) {
  char *rsl;
  FILE *f;

  if (argc != 2) die("usage: get_re xrsl-file");

  f = fopen(argv[1], "r");
  if (!f) die(strerror(errno));

  rsl = read_file(f);
  get_runtimeenvironments(rsl);
  exit(0);
}
