package se.umu.wsag.impl; 

import java.util.Calendar;

import org.apache.log4j.Logger;
import org.ggf.ws_agreement.AgreementPropertiesType;
import org.ggf.ws_agreement.AgreementType;
import org.ggf.ws_agreement.TermTreeType;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.impl.ResourceHomeImpl;
import org.globus.wsrf.impl.SimpleResourceKey;

public class AgreementResourceHome extends ResourceHomeImpl {
	
	private static Logger logger = Logger.getLogger(
			AgreementResourceHome.class); 
	
	public ResourceKey create(AgreementType offer, 
			Object context, 
			AgreementDecisionMaker decisionMaker, 
			AgreementResourceHelper resourceHelper) throws Exception {
		
		try {
			logger.debug("Creating resource"); 
			
			// Set up resource properties: 
			AgreementPropertiesType properties = new AgreementPropertiesType(); 
			properties.setName(offer.getName()); 
			properties.setContext(offer.getContext()); 
			TermTreeType terms = resourceHelper.createTerms(offer, context); 
			if(terms == null)
				throw new Exception("Could not set up agreementTerms in RP"); 
			properties.setTerms(terms); 
			
			AgreementResource resource = (AgreementResource) createNewInstance();	
			
			int resourceKey = properties.hashCode(); 
			if(resourceKey < 0)
				resourceKey *= -1; 
			
			resource.setDecisionMaker(decisionMaker); 
			resource.initialize(properties, 
					//        		null, 
					AgreementPropertiesType.getTypeDesc().getXmlType(), 
					//new Integer(properties.hashCode())); 
					new Integer(resourceKey)); 
			
			// Set initial lifetime as specified by decisionMaker, 
			// default is 1 minute. 
			Calendar cal = Calendar.getInstance(); 
			int lifeTime = decisionMaker.getInitialAgreementLifeTime(); 
			cal.add(Calendar.SECOND, lifeTime);
			resource.setTerminationTime(cal); 
			
			logger.debug("Storing resource"); 
			resource.setDirty(true); 
			resource.store(); 
			
			ResourceKey key = new SimpleResourceKey(keyTypeName, 
					resource.getID());        
			add(key, resource);
			
			
			return key; 
		} catch (Exception e) {
			logger.debug("Bummer!"); 
			throw e; 
		}
		
	}
	
	/**     
	 * Remove all resources (all agreements) when shutting down service
	 * This is not used any more as resources are persistent. 
	 */          
	/* 
	public void destroyAllResources() {
		
		logger.debug("Removing all (" + this.resources.size() + ") resources"); 
		
		Iterator keysIterator = this.resources.keySet().iterator();
		ResourceKey keys[] = new ResourceKey[this.resources.keySet().size()];
		ResourceKey key; 
		
		int i = 0; 
		while(keysIterator.hasNext()) {
			key = (ResourceKey) keysIterator.next(); 
			logger.debug("Key: " + key); 
			keys[i] =  key; 
			i++; 
		}
		
		for(i = 0; i < keys.length; i++) {
			try {
				this.remove(keys[i]);
			}
			catch (Exception e) {
				logger.warn("Error removing resource: " + keys[i]); 
			}    		    	
		}
		
	}
	*/
}
