/*
 * Created on Apr 19, 2005
 *
 *
 */
package se.umu.wsag.impl;

import java.util.Calendar;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.log4j.Logger;
import org.ggf.ws_agreement.AgreementType;
import org.ggf.ws_agreement.GuaranteeTermType;
import org.ggf.ws_agreement.ServiceDescriptionTermType;
import org.ggf.ws_agreement.TermCompositorType;
import org.ggf.ws_agreement.TermTreeType;
import org.globus.delegation.DelegationUtil;
import org.globus.gsi.GlobusCredential;
import org.globus.security.gridmap.GridMap;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.descriptor.ServiceSecurityConfig;
import org.globus.wsrf.impl.security.descriptor.ServiceSecurityDescriptor;
import org.globus.wsrf.security.SecurityManager;

import se.umu.reservation.Config;
import se.umu.reservation.DecisionMakerConfig;
import se.umu.reservation.LocalReservation;
import se.umu.reservation.ReservationDescriptionType;
import se.umu.reservation.ReservationServiceLevelObjective;
import se.umu.reservation.ReservationType;
import se.umu.reservation.StringMap;
import se.umu.wsag.util.Executor;


/**
 * @author tordsson
 *
 * The ReservationDecisionMaker performs admission control
 * against the local (maui) scheduler. It is used by the 
 * AgreementFactoryService before this service creates a
 * new Agreement. The ReservationDecisionMaker is also 
 * used to remove local reservations.  
 * Its configuration is determined by the configuration file
 * for the AgreementFactoryService. 
 */
public class ReservationDecisionMaker implements AgreementDecisionMaker {
			
    private static Logger logger = 
        Logger.getLogger(ReservationDecisionMaker.class);
    
    private String createCommand; 
    private String releaseCommand; 
	private String localMappingClassName; 
	private String mappingContext; 
	
    private int initialAgreementLifeTime; 
    
    private LocalAgreementMapping agreementMapping; 
  
    // Sets up properties specific for the implementation of 
    // DecisionMaker, in this case, the commands used to 
    // create and release reservations.   
    // Also, get the class used to create the local mapping, 
    // in the ARC case, associating the proxy used to create 
    // the reservation with the reservation ID. 
    public void init(Object context) throws Exception {
    	
    	DecisionMakerConfig decisionConfig = (DecisionMakerConfig) context; 
     	Config config = decisionConfig.getConfig(); 
    	
    	createCommand = null; 
    	releaseCommand = null;     	    	
    	
    	this.initialAgreementLifeTime = decisionConfig.
				getInitialAgreementLifeTime(); 
    	if(this.initialAgreementLifeTime == 0)
    		this.initialAgreementLifeTime = 60; 
    	
    	logger.debug("Initial life time is: " + this.initialAgreementLifeTime); 
    	
    	StringMap params[] = config.getContext(); 
        
    	for(int i = 0; i < params.length; i++) {
    		String name = params[i].getName(); 
    		String value = params[i].getValue(); 
    		if("createCommand".equalsIgnoreCase(name))
    			this.createCommand = value; 
    		else if("releaseCommand".equalsIgnoreCase(name))
    			this.releaseCommand = value;
    		else if("localMappingClass".equals(name))
    			this.localMappingClassName = value; 
    		else if("mappingContext".equalsIgnoreCase(name))
    			this.mappingContext = value; 
    		else {
    			logger.warn("Unknown configuration parameter \"" + name + "\""); 
    		}
    	}
    	
    	if(releaseCommand == null || createCommand == null) 
    		throw new Exception("Invalid configuration, must contain both "
    				+ "\"requestCommand\" and  \"releaseCommand\"");     	    	
    	if(localMappingClassName == null || mappingContext == null) 
    		throw new Exception("Invalid configuration, must contain both "
    				+ "\"localMappingClass\" and \"mappingContext\""); 
    	
    	// Set up the localAgreementMapping
    	this.agreementMapping = (LocalAgreementMapping) Class.forName(
    			localMappingClassName).newInstance(); 
    	this.agreementMapping.init(mappingContext); 
    	
    	logger.debug("Local mapping class is: " +this.agreementMapping.getClass()); 
    	
    }
    
    /**
     * @return The configuration of this decision maker, 
     * useful for saving the configuration of the decision maker 
     * on persistant storage (when storing a Reservation Resource). 
     */
    public DecisionMakerConfig getConfig() {
    	
    	DecisionMakerConfig decisionConfig = new DecisionMakerConfig(); 
    	decisionConfig.setClassName(ReservationDecisionMaker.class.getName()); 
    	decisionConfig.setInitialAgreementLifeTime(
    			this.initialAgreementLifeTime); 
    	
    	StringMap createCommand = new StringMap(); 
    	createCommand.setName("createCommand"); 
    	createCommand.setValue(this.createCommand); 
    	StringMap releaseCommand= new StringMap(); 
    	releaseCommand.setName("releaseCommand"); 
    	releaseCommand.setValue(this.releaseCommand);     	
    	
    	StringMap localMappingClass = new StringMap(); 
    	localMappingClass.setName("localMappingClass"); 
    	localMappingClass.setValue(this.localMappingClassName); 
    	StringMap mappingContext = new StringMap(); 
    	mappingContext.setName("mappingContext"); 
    	mappingContext.setValue(this.mappingContext); 
    	
    	StringMap context [] = new StringMap[] {createCommand, 
    			releaseCommand, 
				localMappingClass, 
				mappingContext}; 
    	
    	Config config = new Config(); 
    	config.setContext(context); 
    	
    	decisionConfig.setConfig(config);
    	
    	return decisionConfig; 
    }
	
    /**
     * Extracts reservation-specific parameters from the offer. 
     * @param offer
     * @return
     */
	public ReservationType extractRequest(AgreementType offer) {
		try {
		
			ReservationType reservationOffer = new ReservationType(); 
			
			TermTreeType terms = offer.getTerms(); 
			TermCompositorType all = terms.getAll();
			
			// ServiceDescriptionTerms: 			
			ServiceDescriptionTermType term = all.getServiceDescriptionTerm();				
			
			MessageElement any[] = term.get_any(); 			
			ReservationDescriptionType reservationDescription = 
				(ReservationDescriptionType) ObjectDeserializer.
				toObject(any[0], ReservationDescriptionType.class);
			
			//NumberOfCPUsType noCPUs = (NumberOfCPUsType) ObjectDeserializer.
			//	toObject(any[0], NumberOfCPUsType.class);
			int noCPUs = reservationDescription.getNumberOfCPUs(); 			
			reservationOffer.setNumberOfCPUs(noCPUs);
			logger.debug("Requested: " + noCPUs	+ " CPUs"); 		
			
			boolean flexible = reservationDescription.isFlexible(); 
			reservationOffer.setFlexible(flexible); 
			logger.debug("Reservation is " + (flexible? "" : "not ") 
					+ "flexible"); 
			
			int duration = reservationDescription.getDuration(); 
			reservationOffer.setDuration(duration); 
			logger.debug("Duration: " + duration + " seconds");
			
			// GuaranteeTerms (we skip service properties for now, 
			// these will come in handy when (if) XQuery (or XPath) is used
			// for interpreting offer. 
			
			GuaranteeTermType guaranteeTerms = all.getGuaranteeTerm();
			
			// Get earliest and latest start from reservationSLOs. 
			Object SLOs = guaranteeTerms.getServiceLevelObjective();
			ReservationServiceLevelObjective reservationSLOs; 
			if(SLOs instanceof ReservationServiceLevelObjective) {
				
				reservationSLOs = (ReservationServiceLevelObjective) SLOs; 
				String SLOTerms[] = reservationSLOs.getTerm();
				String SLOTerm = null; 
				
				// FIXME: This parsing is really ugly...								
				for(int i = 0; i < SLOTerms.length; i++) {					
					SLOTerm = SLOTerms[i]; 
					int pos1 = SLOTerm.indexOf(" "); 
					String termName = SLOTerm.substring(0, pos1); 
					int pos2 = SLOTerm.indexOf(" ", pos1 + 1); 
					String termOp = SLOTerm.substring(pos1 + 1, pos2); 
					String termValue = SLOTerm.substring(pos2 + 1, 
							SLOTerm.length());				
					
					if("IS_AFTER".equals(termOp) && 
							"earlyStart".equals(termName)) {
						try {
							Calendar early = Calendar.getInstance();
							long time = Long.parseLong(termValue);						
							early.setTimeInMillis(time*1000); 										
							reservationOffer.setEarliestStart(early); 
									// new EarliestStartType(early));
						} catch (Exception e) {
							reservationOffer.setEarliestStart(null); 
						}						
					}
					if("IS_BEFORE".equals(termOp) &&
							"lateStart".equals(termName)) {
						try {
							Calendar late = Calendar.getInstance(); 
							long time = Long.parseLong(termValue); 
							late.setTimeInMillis(time*1000); 
							reservationOffer.setLatestStart(late); 
									// new LatestStartType(late));
						} catch (Exception e) {
							reservationOffer.setLatestStart(null); 
						}																		
					}	
				}
			}
			else {
				logger.debug("Unknown type for ServiceLevelObjectives!");
				return null; 
			}
			
			return reservationOffer;
		} catch (Exception e) {
			return null; 
		}
			
	}
	
	/**
	 * Retrieves the local user ID for the user executing the service
	 * from the gridmap file. 
	 * @return the local user id
	 */
	public String getLocalUserID() {
		
		String localUser; 
		try {
			SecurityManager manager = SecurityManager.getManager(); 
			String globusID = manager.getCaller(); 

           /* Code used by JobManagerScript.java, try it out: 
              gridMapFile = ContainerSecurityConfig.getConfig().
              	  getSecurityDescriptor().getGridMapFile(); */ 
			
			// Get location of grid-mapfile
			
			ServiceSecurityDescriptor desc = ServiceSecurityConfig.
                            getSecurityDescriptor();        	
			
			String gridMapFileName = desc.getGridMapFile();
			
			if(gridMapFileName == null) {
				logger.warn("Could not find grid map file, using default grid map location"); 
				gridMapFileName = "/etc/grid-security/grid-mapfile"; 
			}
			 
			// Get local userid
			GridMap gridMap = new GridMap(); 
			gridMap.load(gridMapFileName);         	        	
			localUser = gridMap.getUserID(globusID); 
			
			return localUser; 
			
		} catch (Exception e) {
			logger.error("Could not retrieve local id:");
			logger.error("[" + e.getMessage() + "]");         			
			return null; 
		}
	}
	
	/**
	 * This method interfaces the Maui scheduler, and checks if a reservation
	 * can be created
	 */
	public LocalReservation createLocalReservation(int numberOfCPUs, 
			Calendar earliestStart, 
			Calendar latestStart, 
			int duration, 	// Must be in seconds...
			boolean flexible, 
			GlobusCredential delegatedCredential) throws Exception {			
		
		String command = this.createCommand; 
		
	    if(numberOfCPUs <= 0) {
	    	logger.debug("Number of CPUs must be positive"); 
            throw new Exception("Number of CPUs must be positive"); 
        }
	    
	    if(duration <= 0) {
	    	logger.debug("Duration must be positive"); 	    
	    	throw new Exception("Duration must be positive");
	    }
	    
        command += " -n " + numberOfCPUs; 
        command += " -d " + duration; 

        long earliestStartSeconds = earliestStart.getTimeInMillis() / 1000; 
        long latestStartSeconds = latestStart.getTimeInMillis() / 1000; 

        // Earliest and latest job start. 
        command += " -e " + earliestStartSeconds;
        command += " -l " + latestStartSeconds; 
        
        logger.debug("Executing: " + command); 
            
        // Get local userID from gridmap 
        String localUser = null;
        localUser = getLocalUserID(); 
        if(localUser == null) {        	        
        	String gridUser  = SecurityManager.getManager().getCaller(); 
        	logger.debug("Can not find local userID for " + gridUser);
        	throw new Exception("Can not find local userID " + gridUser); 
        }             
        
        // Owner of local reservation == Caller          
        command += " -u " + localUser;  
                
        // Must be placed last
        if(flexible) 
            command += " -f"; 

        // Run command
        String output = null; 
        try {
            output = Executor.runCommand(command); 
        } catch (Exception e) {
            logger.error("Error executing " + command); 
            throw e;
        }

        // Parse output, format is either "failed" or
        // reservationid starttime
        // 'failed' means failed, and will never work.         
        // 'failed 358' means failed, but will (may) work 
        // if requested 358 seconds later... 
        // The WS-Agreement specification mentions that 
        // "A fault response indicates that the offer was rejected 
        // and may also indicate domainspecific reasons."        
        // With that in mind, an Exception containing 
        // "nextstart 358" is returned upon "failed 358" 
        output.trim(); 
        logger.info("Output is: " + output); 
        
        if(output.startsWith("failed")) {
        	// Reservation failed, never possible
        	if(output.equalsIgnoreCase("failed")) {
        		logger.debug("Local reservation could not be granted..."); 
        		throw new Exception("Failed to created reservation...");
        	}
        	else {
        		int pos = output.indexOf(" "); 
        		String nextStart = output.substring(pos + 1).trim(); 
        		logger.debug("Failed, but is possible in another " 
        				+ nextStart + " seconds"); 
        		throw new Exception("nextstart " + nextStart);
        	}
        }

        int pos = output.indexOf(" "); 
        String localID = output.substring(0, pos); 
        String startTimeString = output.substring(pos + 1).trim(); 
        logger.info("Reservation: [" + localID + "] successfully created");
        logger.debug("Exact start time is: " + startTimeString); 
        
        Date startTimeDate = new Date(Long.parseLong(startTimeString)*1000); 

        logger.debug("Exact start as date: " + startTimeDate); 
        Calendar startTime = Calendar.getInstance(); 
        startTime.setTime(startTimeDate);
        
        // Add local mapping for reservation (middleware specific) 
        try {
        	List l = new LinkedList(); 
        	l.add(delegatedCredential); 
        	l.add(localID); 
        	this.agreementMapping.create(l); 
        } catch (Exception e) {
        	e.printStackTrace(); 
        	// Cancel local reservation if local mapping can not be created
        	logger.error("Creating of local reservation mapping failed" 
        			+ " removing reservation " + localID); 
        	String releaseCommand = this.releaseCommand + " " + localID; 
        	Executor.runCommand(releaseCommand); 
        	throw e; 
        }
        
        // Return info about local reservation
        LocalReservation reservation = new LocalReservation(); 
        reservation.setLocalID(localID);
        reservation.setExactStart(startTime); 
        reservation.setCommitted(false); 
        
        return reservation; 		
	}
	
/**
 * Extract reservation specific terms from offer, call local scheduler, 
 * to create local reservation, and return decision. 
 */
	public AgreementDecision create(AgreementType offer) {
		
		ReservationType reservationOffer = extractRequest(offer);
 	
		if(reservationOffer == null) 
			return new AgreementDecision(false, "Could not read offer", null); 
		
		// Hmm, seems to be one level of nestling too much in these stubs...
		int numberOfCPUs = reservationOffer.getNumberOfCPUs(); 
			
		Calendar early = null; 
		Calendar late = null; 
		try {		
			early = reservationOffer.getEarliestStart();
			late = reservationOffer.getLatestStart();
		} catch (Exception e) {
			logger.error("Could not determine preferred start time"); 
			return new AgreementDecision(false, "No start time specified", null); 
		}
		// Duration duration = reservationOffer.getDuration().getDuration();
		int duration = reservationOffer.getDuration(); 
		boolean flexible = reservationOffer.isFlexible(); 
		
		// Get epr to delegated credentials, 
		// it is passed as the agreementinitiator. 
		logger.debug("Getting delegated credentials..."); 
		GlobusCredential delegatedCredential = null; 
		try {
			EndpointReferenceType delegatedCredentialEpr = 
				(EndpointReferenceType) offer.getContext().
				getAgreementInitiator(); 
			delegatedCredential = DelegationUtil.
				getDelegationResource(delegatedCredentialEpr).getCredential(); 
			logger.debug("Retrieving delegated credentials: " + 
					delegatedCredential.getIdentity()); 
		} catch (Exception e) {
			logger.error("Could not get delegated credential: " 
					+ e.getMessage()); 
			return new AgreementDecision(false, e.getMessage(), null); 		
		}
		
		logger.debug("Requesting local reservation");
		LocalReservation localReservation = null; 
		try {
			localReservation = createLocalReservation(numberOfCPUs, 
				early, 
				late, 
				duration, 
				flexible, 
				delegatedCredential);
		} catch (Exception e) {
			logger.error("Local reservation failed: " + e.getMessage()); 
			return new AgreementDecision(false, e.getMessage(), null); 		
		}

	    // Pass localID and exactStart as context in AgreementDecision
		return new AgreementDecision(true, "", localReservation); 
	
	}
	
	// cancel, remove the local reservation. 
	public void cancel(ServiceDescriptionTermType terms) {
		
		String localID = null; 		
		try {
			MessageElement any[] = terms.get_any(); 
			ReservationType reservationInfo = (ReservationType) 
				ObjectDeserializer.toObject(any[0], ReservationType.class); 
			localID =  reservationInfo.getLocalID(); 
			logger.debug("Releasing reservation: " + localID); 
			String command = this.releaseCommand + " " + localID;
			logger.debug("Executing command [" + command + "]"); 
			Executor.runCommand(command);
			
		} catch (Exception e) {
			logger.error("Error releasing reservation " + localID); 
		}
		try {
			this.agreementMapping.cancel(localID); 
		} catch (Exception e) {
			logger.error("Error removing local proxy file for reservation: " 
					+ localID); 
		}
			
	}

	/**
	 * @return Returns the initialAgreementLifeTime.
	 */
	public int getInitialAgreementLifeTime() {
		return initialAgreementLifeTime;
	}
}
